package jalview.datamodel.features;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

import jalview.datamodel.SequenceFeature;

import java.util.List;

import org.testng.annotations.Test;

public class FeatureStoreTest
{

  @Test(groups = "Functional")
  public void testFindFeatures_nonNested()
  {
    FeatureStore fs = new FeatureStore();
    fs.addFeature(new SequenceFeature("", "", 10, 20, Float.NaN,
            null));
    fs.addFeature(new SequenceFeature("", "", 10, 20, Float.NaN, null));
    fs.addFeature(new SequenceFeature("", "", 15, 25, Float.NaN, null));
    fs.addFeature(new SequenceFeature("", "", 20, 35, Float.NaN, null));

    List<SequenceFeature> overlaps = fs.findOverlappingFeatures(1, 9);
    assertTrue(overlaps.isEmpty());

    overlaps = fs.findOverlappingFeatures(8, 10);
    assertEquals(overlaps.size(), 2);
    assertEquals(overlaps.get(0).getEnd(), 20);
    assertEquals(overlaps.get(1).getEnd(), 20);

    overlaps = fs.findOverlappingFeatures(12, 16);
    assertEquals(overlaps.size(), 3);
    assertEquals(overlaps.get(0).getEnd(), 20);
    assertEquals(overlaps.get(1).getEnd(), 20);
    assertEquals(overlaps.get(2).getEnd(), 25);

    overlaps = fs.findOverlappingFeatures(33, 33);
    assertEquals(overlaps.size(), 1);
    assertEquals(overlaps.get(0).getEnd(), 35);
  }

  @Test(groups = "Functional")
  public void testFindFeatures_nested()
  {
    FeatureStore fs = new FeatureStore();
    SequenceFeature sf1 = addFeature(fs, 10, 50);
    SequenceFeature sf2 = addFeature(fs, 10, 40);
    SequenceFeature sf3 = addFeature(fs, 20, 30);
    SequenceFeature sf4 = addFeature(fs, 20, 30);
    SequenceFeature sf5 = addFeature(fs, 35, 36);

    List<SequenceFeature> overlaps = fs.findOverlappingFeatures(1, 9);
    assertTrue(overlaps.isEmpty());

    overlaps = fs.findOverlappingFeatures(10, 15);
    assertEquals(overlaps.size(), 2);
    assertTrue(overlaps.contains(sf1));
    assertTrue(overlaps.contains(sf2));

    overlaps = fs.findOverlappingFeatures(45, 60);
    assertEquals(overlaps.size(), 1);
    assertTrue(overlaps.contains(sf1));

    overlaps = fs.findOverlappingFeatures(32, 38);
    assertEquals(overlaps.size(), 3);
    assertTrue(overlaps.contains(sf1));
    assertTrue(overlaps.contains(sf2));
    assertTrue(overlaps.contains(sf5));

    overlaps = fs.findOverlappingFeatures(15, 25);
    assertEquals(overlaps.size(), 4);
    assertTrue(overlaps.contains(sf1));
    assertTrue(overlaps.contains(sf2));
    assertTrue(overlaps.contains(sf3));
    assertTrue(overlaps.contains(sf4));
  }

  @Test(groups = "Functional")
  public void testFindFeatures_mixed()
  {
    FeatureStore fs = new FeatureStore();
    SequenceFeature sf1 = addFeature(fs, 10, 50);
    SequenceFeature sf2 = addFeature(fs, 1, 15);
    SequenceFeature sf3 = addFeature(fs, 20, 30);
    SequenceFeature sf4 = addFeature(fs, 40, 100);
    SequenceFeature sf5 = addFeature(fs, 60, 100);
    SequenceFeature sf6 = addFeature(fs, 70, 70);

    List<SequenceFeature> overlaps = fs.findOverlappingFeatures(200, 200);
    assertTrue(overlaps.isEmpty());

    overlaps = fs.findOverlappingFeatures(1, 9);
    assertEquals(overlaps.size(), 1);
    assertTrue(overlaps.contains(sf2));

    overlaps = fs.findOverlappingFeatures(5, 18);
    assertEquals(overlaps.size(), 2);
    assertTrue(overlaps.contains(sf1));
    assertTrue(overlaps.contains(sf2));

    overlaps = fs.findOverlappingFeatures(30, 40);
    assertEquals(overlaps.size(), 3);
    assertTrue(overlaps.contains(sf1));
    assertTrue(overlaps.contains(sf3));
    assertTrue(overlaps.contains(sf4));

    overlaps = fs.findOverlappingFeatures(80, 90);
    assertEquals(overlaps.size(), 2);
    assertTrue(overlaps.contains(sf4));
    assertTrue(overlaps.contains(sf5));

    overlaps = fs.findOverlappingFeatures(68, 70);
    assertEquals(overlaps.size(), 3);
    assertTrue(overlaps.contains(sf4));
    assertTrue(overlaps.contains(sf5));
    assertTrue(overlaps.contains(sf6));
  }

  /**
   * Helper method to add a feature of no particular type
   * 
   * @param fs
   * @param from
   * @param to
   * @return
   */
  SequenceFeature addFeature(FeatureStore fs, int from, int to)
  {
    SequenceFeature sf1 = new SequenceFeature("", "", from, to, Float.NaN,
            null);
    fs.addFeature(sf1);
    return sf1;
  }

  @Test(groups = "Functional")
  public void testFindFeatures_contactFeatures()
  {
    FeatureStore fs = new FeatureStore();

    SequenceFeature sf = new SequenceFeature("disulphide bond", "bond", 10,
            20, Float.NaN, null);
    fs.addFeature(sf);

    /*
     * neither contact point in range
     */
    List<SequenceFeature> overlaps = fs.findOverlappingFeatures(1, 9);
    assertTrue(overlaps.isEmpty());

    /*
     * neither contact point in range
     */
    overlaps = fs.findOverlappingFeatures(11, 19);
    assertTrue(overlaps.isEmpty());

    /*
     * first contact point in range
     */
    overlaps = fs.findOverlappingFeatures(5, 15);
    assertEquals(overlaps.size(), 1);
    assertTrue(overlaps.contains(sf));

    /*
     * second contact point in range
     */
    overlaps = fs.findOverlappingFeatures(15, 25);
    assertEquals(overlaps.size(), 1);
    assertTrue(overlaps.contains(sf));

    /*
     * both contact points in range
     */
    overlaps = fs.findOverlappingFeatures(5, 25);
    assertEquals(overlaps.size(), 1);
    assertTrue(overlaps.contains(sf));
  }

  /**
   * Tests for the method that returns false for an attempt to add a feature
   * that would enclose, or be enclosed by, another feature
   */
  @Test(groups = "Functional")
  public void testAddNonNestedFeature()
  {
    FeatureStore fs = new FeatureStore();

    String type = "Domain";
    SequenceFeature sf1 = new SequenceFeature(type, type, 10, 20,
            Float.NaN, null);
    assertTrue(fs.addNonNestedFeature(sf1));

    // co-located feature is ok
    SequenceFeature sf2 = new SequenceFeature(type, type, 10, 20,
            Float.NaN, null);
    assertTrue(fs.addNonNestedFeature(sf2));

    // overlap left is ok
    SequenceFeature sf3 = new SequenceFeature(type, type, 5, 15, Float.NaN,
            null);
    assertTrue(fs.addNonNestedFeature(sf3));

    // overlap right is ok
    SequenceFeature sf4 = new SequenceFeature(type, type, 15, 25,
            Float.NaN, null);
    assertTrue(fs.addNonNestedFeature(sf4));

    // add enclosing feature is not ok
    SequenceFeature sf5 = new SequenceFeature(type, type, 10, 21,
            Float.NaN, null);
    assertFalse(fs.addNonNestedFeature(sf5));
    SequenceFeature sf6 = new SequenceFeature(type, type, 4, 15, Float.NaN,
            null);
    assertFalse(fs.addNonNestedFeature(sf6));
    SequenceFeature sf7 = new SequenceFeature(type, type, 1, 50, Float.NaN,
            null);
    assertFalse(fs.addNonNestedFeature(sf7));

    // add enclosed feature is not ok
    SequenceFeature sf8 = new SequenceFeature(type, type, 10, 19,
            Float.NaN, null);
    assertFalse(fs.addNonNestedFeature(sf8));
    SequenceFeature sf9 = new SequenceFeature(type, type, 16, 25,
            Float.NaN, null);
    assertFalse(fs.addNonNestedFeature(sf9));
    SequenceFeature sf10 = new SequenceFeature(type, type, 7, 7, Float.NaN,
            null);
    assertFalse(fs.addNonNestedFeature(sf10));
  }
}
